//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
//  File	: osearch.c
//  Usage	: Parameter-based object prototype searches.
//  Author	: Brian Langenfeld (Cipher)
//  Created	: May 14th, 1998
//  Last update : September 20th, 1998
//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

#include "conf.h"
#include "sysdep.h"
#include "structs.h"
#include "utils.h"
#include "comm.h"
#include "interpreter.h"
#include "handler.h"
#include "db.h"

#define MAX_HEADER2_COLUMNS	10
#define OSEARCH_FORMAT \
"Format : osearch [<range>] [-a <applies>] [-n <name>]\r\n" \
"            [-p <position>] [-t <type>]\r\n"

#define VALID_APPLIES \
"   str       class     hit       damroll   illegal\r\n" \
"   dex       level     move      para      combat\r\n" \
"   int       age       gold      rod       stats\r\n" \
"   wis       weight    exp       petri     points\r\n" \
"   con       height    ac        breath\r\n" \
"   cha       mana      hitroll   spell\r\n"

extern struct obj_data *obj_proto;
extern struct index_data *obj_index;
extern int top_of_objt;
extern char *item_types[];
extern char *wear_bits[];
ACMD( oSearch );
bool applyColumns( int N, int applyBit, char *sBuf );

const char *sHeader2Bits[] = {
  "NIL", "Str", "Dex", "Int", "Wis", "Con",	// 00-05
  "Cha", "Cla", "Lev", "Age", "Wei",		// 06-10
  "Hei", "MMP", "MHP", "MVP", "Gol",		// 11-15
  "Exp", " AC", "Hit", "Dam", "SPa",		// 16-20
  "SRo", "SPe", "SBr", "SSp"			// 21-24
};

ACMD( oSearch )
{
  bool bName, bPosition, bType, bApplies, bMet, bRanged;
  char sApplies[MAX_INPUT_LENGTH], sHeader[MAX_STRING_LENGTH], 
	sHeader2[MAX_STRING_LENGTH], sName[MAX_INPUT_LENGTH],
	sObjList[MAX_STRING_LENGTH], sPosition[MAX_INPUT_LENGTH],
	sType[MAX_STRING_LENGTH], cFlag;
  int iFound = 0, iList, iPosition = 0, iType = 0, applyNum,
	iHigh = 32000, iLow = 0, iSwap;
  long lApplies = 0;

  bName = bPosition = bType = bApplies = bMet = bRanged = FALSE;
  sApplies[0] = sHeader[0] = sHeader2[0] = sName[0] = sObjList[0] =
        sPosition[0] = sType[0] = '\0';

  skip_spaces( &argument );
  strcpy( buf, argument );

  if( !( *argument ) || !( argument ) )
  {
    send_to_char( OSEARCH_FORMAT, ch );
    return;
  }

  while( *buf )
  {
    half_chop( buf, arg, buf1 );
    if( isdigit( *arg ) )
    {
      bRanged = TRUE;
      sscanf( arg, "%d-%d", &iLow, &iHigh );
      strcpy( buf, buf1 );
    }
    else if( *arg == '-' )
    {
      cFlag = *(arg + 1);
      switch( cFlag )
      {
        case 'a':
          bApplies = TRUE;
	  half_chop( buf1, sApplies, buf );
	  if( strlen( sApplies ) < 1 )
	  {
	    send_to_char( "The -a switch requires arguments.  Valid arguments are:\r\n", ch );
	    send_to_char( VALID_APPLIES, ch );
	    return;
	  }	
          while( isalpha( *sApplies ) )
          {
            if( !strcmp( sApplies, "str" ) )
              SET_BIT( lApplies, ( 1 << APPLY_STR ) );
            else if( !strcmp( sApplies, "dex" ) )
              SET_BIT( lApplies, ( 1 << APPLY_DEX ) );
            else if( !strcmp( sApplies, "int" ) )
              SET_BIT( lApplies, ( 1 << APPLY_INT ) );
            else if( !strcmp( sApplies, "wis" ) )
              SET_BIT( lApplies, ( 1 << APPLY_WIS ) );
            else if( !strcmp( sApplies, "con" ) )
              SET_BIT( lApplies, ( 1 << APPLY_CON ) );
            else if( !strcmp( sApplies, "cha" ) )
              SET_BIT( lApplies, ( 1 << APPLY_CHA ) );
            else if( !strcmp( sApplies, "class" ) )
              SET_BIT( lApplies, ( 1 << APPLY_CLASS ) );
            else if( !strcmp( sApplies, "level" ) )
              SET_BIT( lApplies, ( 1 << APPLY_LEVEL ) );
            else if( !strcmp( sApplies, "age" ) )
              SET_BIT( lApplies, ( 1 << APPLY_AGE ) );
            else if( !strcmp( sApplies, "weight" ) )
              SET_BIT( lApplies, ( 1 << APPLY_CHAR_WEIGHT ) );
            else if( !strcmp( sApplies, "height" ) )
              SET_BIT( lApplies, ( 1 << APPLY_CHAR_HEIGHT ) );
            else if( !strcmp( sApplies, "mana" ) )
              SET_BIT( lApplies, ( 1 << APPLY_MANA ) );
            else if( !strcmp( sApplies, "hit" ) )
              SET_BIT( lApplies, ( 1 << APPLY_HIT ) );
            else if( !strcmp( sApplies, "move" ) )
              SET_BIT( lApplies, ( 1 << APPLY_MOVE ) );
            else if( !strcmp( sApplies, "gold" ) )
              SET_BIT( lApplies, ( 1 << APPLY_GOLD ) );
            else if( !strcmp( sApplies, "exp" ) )
              SET_BIT( lApplies, ( 1 << APPLY_EXP ) );
            else if( !strcmp( sApplies, "ac" ) )
              SET_BIT( lApplies, ( 1 << APPLY_AC ) );
            else if( !strcmp( sApplies, "hitroll" ) )
              SET_BIT( lApplies, ( 1 << APPLY_HITROLL ) );
            else if( !strcmp( sApplies, "damroll" ) )
              SET_BIT( lApplies, ( 1 << APPLY_DAMROLL ) );
            else if( !strcmp( sApplies, "para" ) )
              SET_BIT( lApplies, ( 1 << APPLY_SAVING_PARA ) );
            else if( !strcmp( sApplies, "rod" ) )
              SET_BIT( lApplies, ( 1 << APPLY_SAVING_ROD ) );
            else if( !strcmp( sApplies, "petri" ) )
              SET_BIT( lApplies, ( 1 << APPLY_SAVING_PETRI ) );
            else if( !strcmp( sApplies, "breath" ) )
              SET_BIT( lApplies, ( 1 << APPLY_SAVING_BREATH ) );
            else if( !strcmp( sApplies, "spell" ) )
              SET_BIT( lApplies, ( 1 << APPLY_SAVING_SPELL ) );
	    else if( !strcmp( sApplies, "combat" ) )
	    {
              SET_BIT( lApplies, ( 1 << APPLY_STR ) );
              SET_BIT( lApplies, ( 1 << APPLY_DEX ) );
              SET_BIT( lApplies, ( 1 << APPLY_AC ) );
	      SET_BIT( lApplies, ( 1 << APPLY_HITROLL ) );
              SET_BIT( lApplies, ( 1 << APPLY_DAMROLL ) );
	    }
	    else if( !strcmp( sApplies, "illegal" ) )
	    {
              SET_BIT( lApplies, ( 1 << APPLY_CLASS ) );
              SET_BIT( lApplies, ( 1 << APPLY_EXP ) );
              SET_BIT( lApplies, ( 1 << APPLY_GOLD ) );
              SET_BIT( lApplies, ( 1 << APPLY_LEVEL ) );
	    }
            else if( !strcmp( sApplies, "points" ) )
            {
              SET_BIT( lApplies, ( 1 << APPLY_MANA ) );
              SET_BIT( lApplies, ( 1 << APPLY_HIT ) );
              SET_BIT( lApplies, ( 1 << APPLY_MOVE ) );
            }
            else if( !strcmp( sApplies, "saves" ) )
            {
              SET_BIT( lApplies, ( 1 << APPLY_SAVING_BREATH ) );
              SET_BIT( lApplies, ( 1 << APPLY_SAVING_PARA ) );
              SET_BIT( lApplies, ( 1 << APPLY_SAVING_PETRI ) );
              SET_BIT( lApplies, ( 1 << APPLY_SAVING_ROD ) );
              SET_BIT( lApplies, ( 1 << APPLY_SAVING_SPELL ) );
            }
	    else if( !strcmp( sApplies, "stats" ) )
	    {
	      SET_BIT( lApplies, ( 1 << APPLY_STR ) );
              SET_BIT( lApplies, ( 1 << APPLY_DEX ) );
              SET_BIT( lApplies, ( 1 << APPLY_INT ) );
              SET_BIT( lApplies, ( 1 << APPLY_WIS ) );
              SET_BIT( lApplies, ( 1 << APPLY_CON ) );
              SET_BIT( lApplies, ( 1 << APPLY_CHA ) );
	    }
  	    else
	    {
	      send_to_char( "You have declared an invalid apply.  Valid applies are:\r\n", ch );
              send_to_char( VALID_APPLIES, ch );
              return;
            }
	    half_chop( buf, sApplies, buf );
          }
	  strcpy( buf1, buf );
	  strcpy( buf, sApplies );
          if( *buf )
   	    strcat( buf, " " );
	  strcat( buf, buf1 );
          break;
        case 'n':
          bName = TRUE;
          half_chop( buf1, sName, buf );
          break;
        case 'p':
          bPosition = TRUE;
          half_chop( buf1, sPosition, buf );
          break;
        case 't':
          bType = TRUE;
          half_chop( buf1, sType, buf );
          break;
        default:
          send_to_char( OSEARCH_FORMAT, ch );
          return;
      }
    }
    else
    {
      send_to_char( OSEARCH_FORMAT, ch );
      return;
    }
  }

  if( bRanged && !( bName || bPosition || bType || bApplies ) )
  {
    send_to_char( "Ranged searches require further parameters.\r\n", ch );
    return;
  }

  if( ( bName && strlen( sName ) < 1 ) ||
        ( bPosition && strlen( sPosition ) < 1 ) ||
        ( bType && strlen( sType ) < 1 ) )
  {
    send_to_char( OSEARCH_FORMAT, ch );
    return;
  }

  // Time to build the header.

  if( bApplies )
    sprintbit( lApplies, sHeader2Bits, sHeader2 );

  sprintf( sHeader, "       vnum    " );
  if( strlen( sHeader2 ) <= ( MAX_HEADER2_COLUMNS * 4 ) )
    strcat( sHeader, "      object name       " );

  strcat( sHeader, sHeader2 );
  strcat( sHeader, "\r\n" );

  if( bPosition )
  {
    for( iPosition = 0; *wear_bits[iPosition] != '\n'; iPosition++ )
      if( isname( sPosition, wear_bits[iPosition] ) )
        break;
    if( *wear_bits[iPosition] == '\n' )
    {
      send_to_char( "You have specified an invalid position.\r\n", ch );
      return;
    }
    else
      iPosition = ( 1 << iPosition );
  }

  if( bType )
  {
    for( iType = 0; *item_types[iType] != '\n'; iType++ )
      if( isname( sType, item_types[iType] ) )
        break;
    if( *item_types[iType] == '\n' )
    {
      send_to_char( "You have specified an invalid item type.\r\n", ch );
      return;
    }
  }

  if( bRanged )
  {
    if( iLow > iHigh )
    {
      iSwap = iLow;
      iLow = iHigh;
      iHigh = iSwap;
    }
  }
  for( iList = 0; iList <= top_of_objt; iList++ )
  {
    if( ( obj_index[iList].vnum < iLow ) || ( obj_index[iList].vnum > iHigh ) )
      continue;

    buf[0] = buf2[0] = '\0';
    bMet = FALSE;

    //------------------------------------------------------------------------------------
    // Name, position, and type are ALWAYS exclusive if declared as a search parameter.
    // If we've specified one of these criteria and the object failed, boot it and move on.
    // Otherwise, mark the object as good to go and keep going.

    if( bName )
    {
      if( isname( sName, obj_proto[iList].name ) )
        bMet = TRUE;
      else
        continue;
    }
    if( bPosition )
    {
      if( CAN_WEAR( &obj_proto[iList], iPosition ) )
        bMet = TRUE;
      else
        continue;
    }
    if( bType )
    {
      if( obj_proto[iList].obj_flags.type_flag == iType )
        bMet = TRUE;
      else
        continue;
    }

    // The base upon which the rest of the information buffer is built.
    sprintf( buf, "%3d. [%5d] ",
        ( iFound + 1 ), obj_index[iList].vnum );

    // If we have less than MAX_HEADER2_COLUMNS columns to display, add the name.
    if( strlen( sHeader2 ) <= ( MAX_HEADER2_COLUMNS * 4 ) )
    {
      sprintf( buf2, "%-25.25s", obj_proto[iList].short_description );
      strcat( buf, buf2 );
    }

    // Now for the apply columns.
    for( applyNum = 1; applyNum < TOP_APPLY_NUM; applyNum++ )
    {
      if( !IS_SET( lApplies, ( 1 << applyNum ) ) )
        continue;
      if( applyColumns( iList, applyNum, buf ) )
        bMet = TRUE;
    }
    strcat( buf, "\r\n" );

    if( !bMet )
      continue;
    if( ( strlen( sObjList ) + ( strlen( buf ) * 2 ) ) > MAX_STRING_LENGTH )
    {
      strcat( sObjList, "Too many matches; try limiting your search or setting a vnum range.\r\n" );
      break;
    }
    else
      strcat( sObjList, buf );
    iFound++;
  }

  if( iFound )
    page_string( ch->desc, sObjList, 1, sHeader );
  else
    send_to_char( "No objects found with specified parameters.\r\n", ch );
}

bool applyColumns( int N, int applyBit, char *sBuf )
{
  if( IS_SET( obj_proto[N].obj_flags.lApplyBits, ( 1 << applyBit ) ) )
  {
    sprintf( buf2, "%4d", obj_proto[N].obj_flags.iApplyMods[applyBit] );
    strcat( sBuf, buf2 );
    return( TRUE );
  }
  strcat( sBuf, "   ." );
  return( FALSE );
}
